import re
import math
from scipy.optimize import minimize
from dateutil.relativedelta import relativedelta
from datetime import datetime, timedelta
import logging
import os
import pathlib

class Support ():

    @staticmethod
    def setup_logger():
        desktop = Support.get_folder_path()
        _logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        _fileHandler = logging.FileHandler(f"{desktop}/support.log", mode='w', encoding='utf-8')
        _fileHandler.setFormatter(_logFormatter)
        logger.addHandler(_fileHandler)
        return logger, _fileHandler
    
    @staticmethod
    def rateConversion(rate):
        if type(rate) == str and re.search('%',rate)!=None:
            return float(rate.replace('%',''))/100
        elif float(rate) > 1:
            return float(rate) / 100
        else:
            return float(rate)
    
    @staticmethod
    def periodicityConversion(per, mode):
        modes = {
            'normal': {'anual': 1, 'a': 1, '6': 1, 'vencimiento': 1, 'v': 1, '7': 1, 'semestral': 2, 's': 2, '5': 2,
                       'cuatrimestral': 3, 'c': 3, '4': 3, 'trimestral': 4, 't': 4, '3': 4, 'bimensual': 6, 'b': 6, '2': 6,
                       'mensual': 12, 'm': 12, '1': 12},
            'amortization': {'anual': 12, 'a': 12, '6': 12, 'vencimiento': 12, 'v': 12, '7': 12, 'semestral': 6, 's': 6, '5': 6,
                             'cuatrimestral': 4, 'c': 4, '4': 4, 'trimestral': 3, 't': 3, '3': 3, 'bimensual': 2, 'b': 2, '2': 2,
                             'mensual': 1, 'm': 1, '1': 1}
        }
        return modes[mode].get(per, 1)
    
    @staticmethod
    def monthSelection(mes):
        months = ['enero', 'febrero', 'marzo', 'abril', 'mayo', 'junio', 'julio', 'agosto', 'septiembre', 'octubre', 'noviembre', 'diciembre']
        return months[mes-1] if 1 <= mes <= 12 else None
    
    @staticmethod
    def years(issuance_date, maturity_date):

        '''def is_leap_year(year):
            """Check if a year is a leap year."""
            return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)
        
        def days_in_year(year):
            """Return the number of days in a year, considering leap years."""
            return 366 if is_leap_year(year) else 365
        
        # Calculate the initial year difference
        year_difference = maturity_date.year - issuance_date.year

        # Adjust the start date by the year difference to see if we've gone too far
        adjusted_date = issuance_date.replace(year=issuance_date.year + year_difference)
        
        # If we've gone too far, decrease the year difference
        if adjusted_date > maturity_date:
            year_difference -= 1
            adjusted_date = issuance_date.replace(year=issuance_date.year + year_difference)
        
        # Calculate the remaining days after the last full year
        remaining_days = (maturity_date - adjusted_date).days
        
        # Calculate the fraction of the year for the remaining days
        # This requires knowing if the last part of the period is a leap year
        final_year_days = days_in_year(adjusted_date.year)
        year_fraction = remaining_days / final_year_days
        
        # Calculate the precise year difference
        precise_year_difference = year_difference + year_fraction
    
        return precise_year_difference'''

        # Define the dates
        start_date = issuance_date
        end_date = maturity_date

        # Calculate the number of complete years
        years = end_date.year - start_date.year
        if (end_date.month, end_date.day) < (start_date.month, start_date.day):
            years -= 1

        # Calculate the number of complete months within the partial year
        months = end_date.month - start_date.month
        if end_date.day < start_date.day:
            months -= 1
        if months < 0:
            months += 12

        # Calculate the remaining days within the partial month
        if end_date.day >= start_date.day:
            days = end_date.day - start_date.day
        else:
            previous_month_end_date = end_date.replace(day=1) - timedelta(days=1)
            days = (previous_month_end_date.day - start_date.day) + end_date.day

        # Calculate the total difference in fractional years
        total_years = years + (months / 12)

        # Check if the end year is a leap year to adjust the days difference
        is_leap_year = lambda year: (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0)
        days_in_year = 366 if is_leap_year(end_date.year) else 365
        total_years += days / days_in_year

        return total_years
    
    @staticmethod
    def get_date_years_ago(end_date, year_difference):
        def is_leap_year(year):
            """Check if a year is a leap year."""
            return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)
        # Calculate the start year by subtracting the truncated year difference from the end year
        start_year = end_date.year - math.trunc(year_difference)
        
        # Handle the case where subtracting the years results in a date before February 29 in a leap year
        if end_date.month == 2 and end_date.day == 29 and not is_leap_year(start_year):
            # Adjust the start date to February 28 if the start year is not a leap year
            start_date = datetime(start_year, end_date.month, 28)
        else:
            # Ensure the new date does not exceed the month's day limit
            try:
                start_date = datetime(start_year, end_date.month, end_date.day)
            except ValueError:
                # Adjust for cases like April 31st; set to the last day of the month
                start_date = datetime(start_year, end_date.month + 1, 1) - timedelta(days=1)
        return start_date.date()
    
    @staticmethod
    def ln_nper(issuance, maturity, fper, type, a_maturity = False):
        if not a_maturity:
            rel_nper = (relativedelta(maturity, issuance).years * 12 / (12 / fper)) + (relativedelta(maturity, issuance).months / (12 / fper))
            if (rel_nper - math.trunc(rel_nper)) > 0 and type == 'fw2': 
                rel_nper += 1
        else:
            rel_nper = 1
        return rel_nper

    @staticmethod
    def days360(start_date,end_date,method_eu=False):
        start_day = start_date.day
        start_month = start_date.month
        start_year = start_date.year
        end_day = end_date.day
        end_month = end_date.month
        end_year = end_date.year
    
        if start_day == 31 or (not method_eu and start_month == 2 and (start_day == 29 or (start_day == 28 and not Support.is_leap_year(start_year)))):
            start_day = 30
    
        if end_day == 31:
            if not method_eu and start_day != 30:
                end_day = 1
                end_month += 1 if end_month != 12 else 1
                end_year += 1 if end_month == 1 else 0
            else:
                end_day = 30
        
        return end_day + end_month * 30 + end_year * 360 - start_day - start_month * 30 - start_year * 360

    @staticmethod
    def solver(v1, v2, v3, v4):
        solution = {}
        objetivo = v4
        
        # Define the objective function
        def objective (x):
            x1, x2, x3, x4 = x
            return (x1 * x2) / x3 

        # Define the constraint function(s)
        def constraint1 (x):
            return ((x[0] * x[1]) / x[2]) - x[3]
        
        # Define the initial guess for the variables
        x0 = [v1, v2, v3, v4]

        # Bounds for the decision variables
        bnds = [(0, None), (v2, v2), (v3, v3), (v4, v4)]
        
        # Constraints: Defines the equality constraint
        cons = [{'type': 'eq', 'fun': constraint1}]
        
        # Tolerance: Sets the termination tolerance for the optimization.
        tolerance = 1e-8

        # Optimization Process: Uses the SLSQP method for constrained optimization.
        sol = minimize(objective, x0, method = 'SLSQP', bounds = bnds, constraints = cons, tol = tolerance)

        # Checking Solution: Verifies if the solution meets the required tolerance and constraints.
        if abs(sol.fun - objetivo) <= tolerance and sol.success and sol.fun > 0:
            t1 = sol.x[0]
            t2 = sol.x[1]
            t3 = sol.x[2]
            test = (t1 * t2) / t3
            solution = {
                'result': t1,
                'test': test,
                'message': f'Solution found within the desired tolerance range with a residual of {test - v4}'
            }
        else:
            solution = {
                'result': 0,
                'test': 0,
                'message': 'No solution found within the desired tolerance range.'
            }
        
        return solution
    
    @staticmethod
    def closestDate(fechas, buscada):
        return min((fecha for fecha in fechas if fecha >= buscada), default=None)
    
    @staticmethod
    def get_folder_path():
        home = pathlib.Path.home()
        desktop = home / "Desktop"
        
        if not desktop.exists():
            desktop = os.path.join(os.path.expanduser("~"), "Desktop")
        
        folder_path = os.path.join(desktop if os.path.exists(desktop) else home, 'cxfiLogs')
        
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
                
        return folder_path
    
    
    
'''
from datetime import datetime, timedelta

# Define the dates
start_date = datetime(2019, 9, 27)
end_date = datetime(2022, 10, 19)

# Calculate the number of complete years
years = end_date.year - start_date.year
if (end_date.month, end_date.day) < (start_date.month, start_date.day):
    years -= 1

# Calculate the number of complete months within the partial year
months = end_date.month - start_date.month
if end_date.day < start_date.day:
    months -= 1
if months < 0:
    months += 12

# Calculate the remaining days within the partial month
if end_date.day >= start_date.day:
    days = end_date.day - start_date.day
else:
    previous_month_end_date = end_date.replace(day=1) - timedelta(days=1)
    days = (previous_month_end_date.day - start_date.day) + end_date.day

# Calculate the total difference in fractional years
total_years = years + (months / 12)

# Check if the end year is a leap year to adjust the days difference
is_leap_year = lambda year: (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0)
days_in_year = 366 if is_leap_year(end_date.year) else 365
total_years += days / days_in_year

total_years

'''


#from tkinter import messagebox
import numpy as np, pandas as pd, math, time, logging
from datetime import date, datetime
from dateutil.relativedelta import relativedelta
from caluxPy_fi.Support import Support

class FixedIncome:
    
    def __init__(self, settlement_date, issuance_date, maturity_date, coupon, ytm, face_value, issuer, methodology, periodicity, payment_type, coupon_type, 
                 forward_date, amortization_start_date, amortization_periods, amortization_periodicity, amortizable_percentage, **kwargs):
        
        self.repo_margin = kwargs.get('repo_margin', '')
        self.repo_rate = kwargs.get('repo_rate', '')
        self.repo_tenure = kwargs.get('repo_tenure', '')
        self.flr_rules = kwargs.get('flr_rules', '')
        #self.flr_ratio = kwargs.get('flr_ratio', '')
        self.date_format = kwargs.get('date_format', '%Y-%m-%d')
        self.multiple = kwargs.get('multiple', False)
        self._lang = kwargs.get('lang', 'eng')             
              
        #region ---> Logger
        
        desktop = Support.get_folder_path()
        
        _logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        _fileHandler = logging.FileHandler("{0}/{1}.log".format(desktop, 'fixed_income'), mode = 'w', encoding = 'utf-8')
        _fileHandler.setFormatter(_logFormatter)
        self._logger.addHandler(_fileHandler)

        #endregion ---> Logger
        
        #FixedIncome.counter += 1
        self._logger.info(f'Starting process with Security => Single Calculation') #{FixedIncome.counter}')
        self.results = {}

        self._start_time = time.perf_counter()

        # Date handling
        self.issuance_date = self.ensure_date_type(issuance_date)
        self.maturity_date = self.ensure_date_type(maturity_date)
        self.settlement_date = self.ensure_date_type(settlement_date)
        
        self.issuer = issuer
        
        try:
            self.coupon = Support.rateConversion(rate = coupon)
        except Exception as e:
            self._logger.error(f'Error in Coupon Rate Conversion => {e}')
            
        try:
            self.ytm = Support.rateConversion(rate = ytm)
        except Exception as e:
            self._logger.error(f'Error in YTM Rate Conversion => {e}')
        
        self.payment_type = payment_type
        self.methodology = methodology
        self.face_value = float(face_value)
        self.notional_value = self.face_value
        
        try:
            if self.maturity_date and self.settlement_date and self.maturity_date >= self.settlement_date:
                try:
                    self.periodicity = Support.periodicityConversion(per = periodicity, mode = 'normal')
                except Exception as e:
                    self._logger.error(f'Error in Periodicity Conversion => {e}')
                
                self.a_maturity = True if periodicity in ['vencimiento', 'maturity'] else False
                self.coupon_type = coupon_type
                self._years_to_maturity = Support.years(self.issuance_date, self.maturity_date)

                self.residual_tenure = self.maturity_date - self.settlement_date #this is the remainder days to maturity
                self._total_coupons = self.periodicity * self._years_to_maturity #total amount of coupon payments of the security

                #Determinación del nper, esta parte determina la cantidad de flujos, si es a vencimiento será 1 sólo flujo
                try:
                    self._nper = math.trunc(Support.ln_nper(self.issuance_date, self.maturity_date, self.periodicity, self.coupon_type, self.a_maturity)) - 1
                except Exception as e: 
                    self._logger.error(f'Error determining nper => {e}')
                
                #The numeric periodicicy of the coupon payment, if it is to maturity it will always be equal to 1, but should not be equal to 0
                self._per = (12 / self.periodicity) if self.a_maturity == False else 1 
                
                #This is the expected maturity date of the security and should be the same as the provided, if not a discrepancy will be calculated and taken into account
                try:
                    self.expected_maturity = self.expectedMaturity(maturity_date = self.maturity_date, 
                                                                   issuance_date = self.issuance_date, 
                                                                   years_to_maturity = self._years_to_maturity,
                                                                   nper = self._per)
                except Exception as e: 
                    self._logger.error(f'Error calculating expected maturity => {e}')

                #This is the determination of the existence of a discrepancy between the maturity date and the expected maturity date
                self._discrepancy = "Yes" if (self.maturity_date.day - self.issuance_date.day) != 0 or self.maturity_date != self.expected_maturity else 'No'
                    
                if self.coupon_type == 'normal':
                    self.forward_date = None
                elif coupon_type in ['fw1', 'fw2']:
                    try:
                        self.forward_date = Support.get_date_years_ago(self.maturity_date, self._years_to_maturity)
                    except Exception as e:
                        self._logger.error(f'Error calculating a forward date => {e}')
                else:
                    self.forward_date = self.validate_and_convert(forward_date, 'Forward date', convert_func=self.ensure_date_type)
                        
                #If the security if of bullet type then any attribute of amortizable is deleted to avoid incongruencies
                if self.payment_type == 'bullet':
                    self.amortization_start_date, self.amortization_periods, self.amortization_periodicity, self.amortizable_percentage = None, None, None, None
                else:
                    # If not bullet then all the attributes are kept and verified
                    self._logger.info(f'Amortization Date -> {amortization_start_date} Type => {type(amortization_start_date)}')
                    
                    # Verification of the Amortization Start Date
                    self.amortization_start_date = self.validate_and_convert(amortization_start_date, 'Amortization start date', convert_func=self.ensure_date_type)
                    
                    # Verification of the Amortization Periods
                    self.amortization_periods = self.validate_and_convert(amortization_periods, 'Amortization periods', convert_func=int)
                    
                    # Verification of the Amortization Periodicity
                    self.amortization_periodicity = self.validate_and_convert(amortization_periodicity, 'Amortization periodicity', convert_func=lambda x: Support.periodicityConversion(per=x, mode='amortization'))

                    # Verification of the Amortization Percentage
                    self.amortizable_percentage = self.validate_and_convert(amortizable_percentage, 'Amortizable percentage', convert_func=lambda x: Support.rateConversion(rate=x))

                #Determination of the dates of coupon payments
                try:
                    self.coupon_dates = np.array(self.couponDates(nper = self._nper, 
                                                                  per = self._per, 
                                                                  issuance_date = self.issuance_date, 
                                                                  maturity_date = self.maturity_date, 
                                                                  coupon_type = self.coupon_type, 
                                                                  forward_date = self.forward_date,
                                                                  a_maturity = self.a_maturity))
                except Exception as e: 
                    self._logger.error(f'Error calculating coupon dates => {e}')

                #Iterates the list of coupon dates searching for which are greater than the settlement date and the first one matching is the current coupon
                self.current_coupon = (self.coupon_dates > self.settlement_date).nonzero()[0][0]
                
                try:
                    self.coupon_days, self.flow_num = self.couponDays_flowNum(nper = self._nper, 
                                                                              coupon_dates = self.coupon_dates, 
                                                                              issuance_date = self.issuance_date)
                except Exception as e: 
                    self._logger.error(f'Error calculating coupon days and flow numbers => {e}')
                
                try:
                    self.accumulated_days = self.accumulatedDays(methodology = self.methodology, 
                                                                 issuance_date = self.issuance_date, 
                                                                 fecha_liq = self.settlement_date, 
                                                                 current_coupon = self.current_coupon, 
                                                                 coupon_dates = self.coupon_dates)
                except Exception as e: 
                    self._logger.error(f'Error calculating accumulated days => {e}')
                
                try:
                    self.days_to_maturity = np.array(self.daysToMaturity(nper = self._nper, 
                                                                         coupon_days = self.coupon_days))
                except Exception as e: 
                    self._logger.error(f'Error calculating days to maturity => {e}')
                
                try:
                    self.calculation_basis = np.array(self.calculationBasis(methodology = self.methodology, 
                                                                            nper = self._nper, 
                                                                            coupon_dates = self.coupon_dates))
                except Exception as e: self._logger.error(f'Error calculating calculation basis => {e}')
                
                try:
                    self.w_coupon = self.wCoupon(methodology = self.methodology, 
                                                 coupon_days = self.coupon_days, 
                                                 current_coupon = self.current_coupon, 
                                                 accumulated_days = self.accumulated_days, 
                                                 periodicity = self.periodicity)
                except Exception as e: 
                    self._logger.error(f'Error calculating w coupon => {e}')
                    
                if self.payment_type in ['amortized', 'amortizado', 'pik']:
                    try:
                        self.notional_value, self.amortization_amount, self.adjusted_value, self.dates_results = self.amortizationModule(self.payment_type, self.amortization_start_date, self.maturity_date, 
                                                                                                                                         self.amortization_periods, self.amortization_periodicity, 
                                                                                                                                         self.amortizable_percentage, self.settlement_date, self.face_value, 
                                                                                                                                         self.coupon_dates, self.current_coupon, self._nper, self.periodicity)
                    except Exception as e: 
                        self._logger.error(f'Error in the amortization module => {e}')
                
                try:
                    self.gained_coupon = self.gainedCoupon(methodology = self.methodology, 
                                                           notional_value = self.notional_value, 
                                                           coupon = self.coupon, 
                                                           periodicity = self.periodicity, 
                                                           coupon_days = self.coupon_days, 
                                                           current_coupon = self.current_coupon, 
                                                           accumulated_days = self.accumulated_days, 
                                                           calculation_basis = self.calculation_basis)
                except Exception as e: 
                    self._logger.error(f'Error calculating gained coupon => {e}')
                    
                self.coupon_flow = []
                
                self._logger.info('SUCCESS: All necessary variables have been calculated! Proceeding with the calculation of the cash flows..')
                try:
                    if (self.payment_type in ['bullet']):
                        self.coupon_flow = np.array(self.bulletPaymentType(face_value = self.face_value, 
                                                                           coupon = self.coupon, 
                                                                           periodicity = self.periodicity, 
                                                                           nper = self._nper, 
                                                                           per = self._per, 
                                                                           methodology = self.methodology, 
                                                                           calculation_basis = self.calculation_basis, 
                                                                           coupon_days = self.coupon_days, 
                                                                           current_coupon = self.current_coupon))
                    elif (self.payment_type in ['amortized', 'amortizado']):
                        self.coupon_flow = np.array(self.amortizedPaymentType(methodology = self.methodology, 
                                                                              current_coupon = self.current_coupon, 
                                                                              dates_results = self.dates_results, 
                                                                              adjusted_value = self.adjusted_value, 
                                                                              coupon = self.coupon, 
                                                                              periodicity = self.periodicity, 
                                                                              amortization_amount = self.amortization_amount, 
                                                                              calculation_basis = self.calculation_basis, 
                                                                              coupon_days = self.coupon_days, 
                                                                              amortizable_percentage = self.amortizable_percentage, 
                                                                              nper = self._nper, 
                                                                              per = self._per))
                    elif (self.payment_type in ['pik']):
                        self.coupon_flow = np.array(self.pikPaymentType(methodology = self.methodology, 
                                                                        current_coupon = self.current_coupon, 
                                                                        dates_results = self.dates_results, 
                                                                        notional_value = self.notional_value, 
                                                                        coupon = self.coupon, 
                                                                        periodicity = self.periodicity, 
                                                                        calculation_basis = self.calculation_basis, 
                                                                        coupon_days = self.coupon_days, 
                                                                        nper = self._nper, 
                                                                        per = self._per))
                    self._logger.info('SUCCESS: Cash Flows calculated! Proceeding to calculate their respective present value..')
                except Exception as e: 
                    self._logger.error(f'Error calculating cashflows => {e}')
                
                try:   
                    valuation_results = self.presentValue(current_coupon = self.current_coupon, 
                                                          nper = self._nper, 
                                                          w_coupon = self.w_coupon, 
                                                          discrepancy = self._discrepancy, 
                                                          coupon_type = self.coupon_type, 
                                                          maturity_date = self.maturity_date, 
                                                          expected_maturity = self.expected_maturity, 
                                                          coupon_days = self.coupon_days, 
                                                          ytm = self.ytm, 
                                                          periodicity = self.periodicity, 
                                                          coupon_flow = self.coupon_flow, 
                                                          payment_type = self.payment_type, 
                                                          issuer = self.issuer, 
                                                          notional_value = self.notional_value, 
                                                          gained_coupon = self.gained_coupon)
                    self._logger.info('SUCCESS: Present value calculated! Proceeding with the results..')
                except Exception as e: 
                    self._logger.error(f'Error calculating present value of cashflows => {e}')
                
                self._final_time = time.perf_counter()
                self._process_time = self._final_time - self._start_time
                
                self.present_values = valuation_results[9]
                #Results of normal calculation
                self.results = self.create_results(valuation_results = valuation_results, status = True)
                self.results['status'] = 'Ok'
                
                #Otras operaciones
                try:
                    if all(value not in ['', None, np.nan] for value in [self.repo_margin, self.repo_rate, self.repo_tenure]):
                        # Convert and assign values
                        self.repo_margin = float(Support.rateConversion(self.repo_margin))
                        self.repo_rate = float(Support.rateConversion(self.repo_rate))
                        self.repo_tenure = int(self.repo_tenure)
                        
                        # Calculate repo results
                        resultado_repos = self.Repos(margin = self.repo_margin,
                                                     repo_rate = self.repo_rate,
                                                     repo_tenure = self.repo_tenure)
                        
                        # Assign results to the results dictionary
                        if self.multiple:
                            self.results['Valor Ida' if self._lang == 'esp' else 'Near Leg'], self.results['Valor Vuelta' if self._lang == 'esp' else 'Far Leg'] = resultado_repos
                        else:
                            self.results['valor_ida' if self._lang == 'esp' else 'near_leg'], self.results['valor_vuelta' if self._lang == 'esp' else 'far_leg'] = resultado_repos
                except ValueError as e:
                    self._logger.error(f'Value error in repo calculation: {e}')
                except TypeError as e:
                    self._logger.error(f'Type error in repo calculation: {e}')
                except Exception as e:
                    self._logger.error(f'Unexpected error calculating repo: {e}')
                
                try:
                    if self.flr_rules:
                        self.available_amount, self.objective_notional = self.FLR(rules = self.flr_rules)
                        if self.multiple:
                            self.results['Disponible' if self._lang == 'esp' else 'Available'] = self.available_amount
                            self.results['Objetivo' if self._lang == 'esp' else 'Objective'] = self.objective_notional
                        else:
                            self.results['monto_disponible' if self._lang == 'esp' else 'available_amount'] = self.available_amount
                            self.results['monto_objetivo' if self._lang == 'esp' else 'objective_amount'] = self.objective_notional

                except Exception as e:
                    self._logger.error(f'Error calculating flr => {e}')
                    
                self._logger.info('SUCCESS: Results ready! Preparing the visualization..')

                values = {}
                est_flujos_exp, flujos_exp, present_values_export, accumulated_days_export = [], [], [], []

                days = 0
                for i in range(len(self.flow_num)):
                    days += self.coupon_days[i]
                    accumulated_days_export.append(days)
                    if i < self.current_coupon:
                        est_flujos_exp.append('Vencido' if self._lang == 'esp' else 'Matured')
                        flujos_exp.append(0)
                        present_values_export.append(0)
                    else:
                        est_flujos_exp.append('Vigente' if self._lang == 'esp' else 'Active')
                        flujos_exp.append(self.coupon_flow[i - self.current_coupon])
                        present_values_export.append(self.present_values[i - self.current_coupon])

                values['Flujo' if self._lang == 'esp' else 'Flow'] = np.array(self.flow_num)
                values['Fecha Cupón' if self._lang == 'esp' else 'Coupon Date'] = np.array(self.coupon_dates)
                values['Días Cupón' if self._lang == 'esp' else 'Coupon Days'] = np.array(self.coupon_days)
                values['Días Acumulados' if self._lang == 'esp' else 'Accumulated Days'] = np.array(accumulated_days_export)
                values['Flujo Cupón' if self._lang == 'esp' else 'Coupon Flow'] = np.array(flujos_exp)
                values['Valor Presente' if self._lang == 'esp' else 'Present Value'] = np.array(present_values_export)
                values['Vigencia' if self._lang == 'esp' else 'Status'] = np.array(est_flujos_exp)

                self.df_results = pd.DataFrame(values)
                self._logger.info('SUCCESS: Results visualization ready! Exiting this procedure..')
            else:
                raise Exception('Security matured.')
        except Exception as e:
            self._logger.exception(e)
            self.results = self.create_results(valuation_results = [], status = False)
            self._logger.error('Título Vencido..' if self._lang == 'esp' else 'Matured Security..')
            self.results['status'] = 'Vencido' if self._lang == 'esp' else 'Matured'
            self._logger.error('Process cancelled because of an error')
        finally:
            self._logger.removeHandler(_fileHandler)
            _fileHandler.close()

    def ensure_date_type(self, input_date):
        ensured_date = None
        
        try:
            if isinstance(input_date, str):
                ensured_date = datetime.strptime(input_date, self.date_format).date()
            elif isinstance(input_date, datetime):
                ensured_date = input_date.date()
            elif isinstance(input_date, pd.Timestamp):
                ensured_date = input_date.date()
            elif isinstance(input_date, date):
                ensured_date = input_date
            else:
                raise TypeError("Unsupported type for date")
        except (ValueError, TypeError) as e:
            self._logger['error'](f"Error processing date: {e}")
        return ensured_date

    def validate_and_convert(self, value, value_name, convert_func=None, logger=True, raise_exception=True):
        """
        Validate and optionally convert a value.

        Args:
            value: The value to validate and convert.
            value_name (str): The name of the value (for logging purposes).
            convert_func (callable, optional): A function to convert the value. Defaults to None.
            logger (logging.Logger, optional): A logger for logging errors. Defaults to None.
            raise_exception (bool, optional): Whether to raise an exception on error. Defaults to True.

        Returns:
            The converted value, or the original value if no conversion function is provided.
        
        Raises:
            ValueError: If the value is invalid and raise_exception is True.
        """
        is_invalid = pd.isna(value) or value in ['', None, pd.NaT]
        
        if is_invalid:
            error_message = f'{value_name} not specified or invalid!'
            if logger:
                self._logger.error(error_message)
            if raise_exception:
                raise ValueError(error_message)
            return None
        
        if convert_func:
            try:
                return convert_func(value)
            except Exception as e:
                error_message = f'Error converting {value_name} => {e}'
                if logger:
                    self._logger.error(error_message)
                if raise_exception:
                    raise ValueError(f'Invalid {value_name} value {value}')
                return None
            
        return value
    
    def expectedMaturity(self, maturity_date, issuance_date, years_to_maturity, nper):
        def check_months(maturity, issuance, nper):
            if maturity.month - issuance.month < 0:
                return maturity.month - issuance.month + nper
            elif maturity.month - issuance.month > 0:
                return maturity.month - issuance.month - nper
        #Determination of the expected maturity date
        if (check_months(maturity_date, issuance_date, nper)) != 0 or maturity_date.day != issuance_date.day:
            expected_maturity = date(issuance_date.year + math.trunc(years_to_maturity),issuance_date.month,issuance_date.day)
        else:
            expected_maturity = maturity_date
        return expected_maturity

    def couponDates(self, nper, per, issuance_date, maturity_date, coupon_type, forward_date, a_maturity = False): 
        #Determination of the dates in which coupons would be paid
        coupon_dates = []
        if a_maturity == False:
            i = 0
            while i <= nper:
                if i == 0:
                    if coupon_type == 'fw1':
                        coupon_dates.append((forward_date + relativedelta(months=+per)))
                    elif (coupon_type in ['fw2', 'fw3']):
                            coupon_dates.append(forward_date)
                    else:
                        coupon_dates.append((issuance_date + relativedelta(months=+per)))
                else:
                    if i == nper:
                        coupon_dates.append(maturity_date)
                    elif (coupon_type in ['fw2', 'fw3']):
                        coupon_dates.append((forward_date + relativedelta(months=+per * i)))
                    else:
                        coupon_dates.append((issuance_date + relativedelta(months=+per*(1+i))))   
                i += 1
        else:
           coupon_dates.append(maturity_date) 
        return coupon_dates

    def couponDays_flowNum(self, nper, coupon_dates, issuance_date): 
        #a) Determination of the days between each coupon
        #b) Determination of the flow number of the coupons
        coupon_days, flow_num = [], []
        i = 0
        while i <= nper:
            flow_num.append(i)
            if i == 0:
                coupon_days.append((coupon_dates[0] - issuance_date).days)
            else:
                coupon_days.append((coupon_dates[i] - coupon_dates[i - 1]).days)
            i += 1
        return coupon_days, flow_num

    def accumulatedDays(self, methodology, issuance_date, fecha_liq, current_coupon, coupon_dates): 
        #Determination of the accumulated days
        if (methodology != "isma-30-360") and current_coupon == 0:
            accumulated_days = (fecha_liq - issuance_date).days
        elif (methodology != "isma-30-360") and current_coupon != 0:
            accumulated_days = (fecha_liq - coupon_dates[current_coupon-1]).days
        elif (methodology == "isma-30-360") and current_coupon == 0:
            accumulated_days = (Support.days360(issuance_date,fecha_liq)).days
        elif (methodology == "isma-30-360") and current_coupon != 0:
            accumulated_days = (Support.days360(coupon_dates[current_coupon-1],fecha_liq))
        return accumulated_days

    def wCoupon(self, methodology, coupon_days, current_coupon, accumulated_days, periodicity): 
        #Determination of the w coupon
        if (methodology == 'isma-30-360'):
            w_coupon = ((360 / periodicity) - accumulated_days) / (360 / periodicity)
        else:
            w_coupon = (coupon_days[current_coupon] - accumulated_days) / coupon_days[current_coupon]
        return w_coupon

    def daysToMaturity(self, nper, coupon_days): 
        #Determination of the days between each coupon and the maturity date
        days_to_maturity = []
        i = 0
        while i <= nper:
            if i == 0:
                days_to_maturity.append(coupon_days[0])
            else:
                days_to_maturity.append(days_to_maturity[i - 1] + coupon_days[i])
            i += 1 
        return days_to_maturity

    def calculationBasis(self, methodology, nper, coupon_dates): 
        #Determination of the calculation basis
        calculation_basis = []
        i = 0
        while i <= nper:
            if (methodology == 'actual/365'): calculation_basis.append(365)
            elif (methodology in ['actual/360', 'isma-30-360']): calculation_basis.append(360)
            else: calculation_basis.append((coupon_dates[i] - (coupon_dates[i] - relativedelta(months=+12))).days)
            i += 1
        return calculation_basis

    def amortizationModule(self, payment_type, amortization_start_date, maturity_date, amortization_periods, amortization_periodicity, amortizable_percentage, 
                           settlement_date, face_value, coupon_dates, current_coupon, nper, periodicity):

        #Determination of the dates in which the securities amortize
        amortization_amount = 0
        amortization_dates = []
        dates_results = []
        if (payment_type in ['amortized', 'amortizado', 'pik']):
            date_count = 0
            if amortization_start_date < settlement_date:
                date_count = 1
            i = 0
            while i <= int(amortization_periods) - 1:
                if i == 0:
                    amortization_dates.append(amortization_start_date)
                else:
                    amortization_dates.append(amortization_dates[i - 1] + relativedelta(months=+amortization_periodicity))
                    if amortization_dates[i] <= settlement_date:
                        date_count += 1
                i += 1
            amortization_dates = np.array(amortization_dates)
            i = 0
            posicion = 0
            while i <= np.count_nonzero(amortization_dates) - 1:
                posicion = np.where(coupon_dates == amortization_dates[i])[0]
                if posicion != 0: dates_results.append(posicion)
                i += 1
            amortization_amount = (face_value * amortizable_percentage) / amortization_periods
            dates_results = np.array(dates_results)
        #Calculating the Notional Value and the Adjusted Amounts
        previously_amortized = 0
        notional_value = 0
        adjusted_value = []
        if (payment_type in ['amortized', 'amortizado']):
            if amortization_periods > 0:
                if np.count_nonzero(dates_results) != date_count:
                    previously_amortized = amortization_amount * date_count
                notional_value = face_value - previously_amortized
            count = 0
            i = current_coupon
            while i <= nper:
                if np.where(dates_results == i)[0].size > 0:
                    count += 1
                    if count <= (amortization_periods - (amortization_periods - np.count_nonzero(dates_results))):
                        adjusted_value.append(notional_value - (amortization_amount * (count - 1)))
                    else:
                        adjusted_value.append(face_value)
                else:
                    adjusted_value.append(notional_value - (amortization_amount * count))
                i += 1
            adjusted_value = np.array(adjusted_value)

        return notional_value, amortization_amount, adjusted_value, dates_results

    def gainedCoupon(self, methodology, notional_value, coupon, periodicity, coupon_days, current_coupon, accumulated_days, calculation_basis): 
        #Calculates the running coupon or gained interest
        gained_coupon = 0
        if (methodology in ['icma']): gained_coupon = (notional_value * coupon / periodicity) / coupon_days[current_coupon] * accumulated_days
        elif (methodology in ['actual/actual', 'actual/365', 'actual/360']): gained_coupon = notional_value * coupon / calculation_basis[current_coupon] * accumulated_days
        elif (methodology in ['isma-30-360']): gained_coupon = notional_value * coupon / 360 * accumulated_days
        return gained_coupon

    def bulletPaymentType(self, face_value, coupon, periodicity, nper, per, methodology, calculation_basis, coupon_days, current_coupon):
        coupon_flow = []
        i = current_coupon
        while i <= nper:
            if (methodology in ['icma']): coupon_flow.append((face_value * coupon) / periodicity)
            elif (methodology in ['isma-30-360']): coupon_flow.append((face_value * coupon) / 12 * per)
            elif (methodology in ['actual/actual', 'actual/365', 'actual/360']): coupon_flow.append(face_value * coupon / calculation_basis[i] * coupon_days[i])
            
            if i == nper: 
                coupon_flow[int(nper - current_coupon)] = face_value + coupon_flow[int(nper - current_coupon)]
            i += 1
        return coupon_flow
    
    def amortizedPaymentType(self, methodology, current_coupon, dates_results, adjusted_value, coupon, periodicity, amortization_amount, calculation_basis, coupon_days, amortizable_percentage, nper, per):
        i = current_coupon
        coupon_flow = []
        x = 0
        while i <= nper:
            if (methodology in ['icma']):
                if np.where(dates_results == i)[0].size > 0: #this coupon amortizes
                    coupon_flow.append((adjusted_value[x] * coupon / periodicity) + amortization_amount)
                else:#normal coupon
                    coupon_flow.append(adjusted_value[x] * coupon / periodicity)
            elif (methodology in ['isma-30-360']):
                if np.where(dates_results == i)[0].size > 0: #this coupon amortizes
                    coupon_flow.append(((adjusted_value[x] * coupon) / 12 * per) + amortization_amount)
                else: #normal coupon
                    coupon_flow.append((adjusted_value[x] * coupon) / 12 * per)
            elif (methodology in ['actual/actual', 'actual/365', 'actual/360']):
                if np.where(dates_results == i)[0].size > 0: #this coupon amortizes
                    coupon_flow.append((adjusted_value[x] * coupon / calculation_basis[i] * coupon_days[i]) + amortization_amount)
                else: #normal coupon
                    coupon_flow.append(adjusted_value[x] * coupon / calculation_basis[i] * coupon_days[i])
            if i == nper:
                if np.where(dates_results == i)[0].size > 0:
                    if amortizable_percentage == 1: #if amortizes 100%
                        coupon_flow[nper - current_coupon] = coupon_flow [x]
                    else: #if does not amortize 100%
                        coupon_flow[x] = adjusted_value[x] + coupon_flow[x] + amortization_amount
                else: 
                    coupon_flow[x] = adjusted_value[x] + coupon_flow[x]
            x += 1
            i += 1
        return coupon_flow
    
    def pikPaymentType(self, methodology, current_coupon, dates_results, notional_value, coupon, periodicity, calculation_basis, coupon_days, nper, per):
        i = current_coupon
        coupon_flow = []
        x = 0
        while i <= nper:
            if (methodology in ['icma']):
                if np.where(dates_results == i)[0].size > 0:
                    if i == current_coupon:
                        coupon_flow.append(notional_value + (notional_value * coupon / periodicity))
                    else:
                        coupon_flow.append(coupon_flow[x - 1] + (coupon_flow[x - 1] * coupon / periodicity))
                else:
                    coupon_flow.append(coupon_flow[x - 1] * coupon / periodicity)
            elif (methodology in ['isma-30-360']):
                if np.where(dates_results == i)[0].size > 0:
                    if i == current_coupon:
                        coupon_flow.append(notional_value + (notional_value * coupon) / 12 * per)
                    else:
                        coupon_flow.append(coupon_flow[x - 1] + ((coupon_flow[x - 1] * coupon) / 12 * per))
                else:
                    coupon_flow.append(coupon_flow[x - 1] * coupon / 12 * per)
            elif (methodology in ['actual/actual', 'actual/365', 'actual/360']):
                if np.where(dates_results == i)[0].size > 0:
                    if i == current_coupon:
                        coupon_flow.append(notional_value + (notional_value * coupon / calculation_basis[i] * coupon_days[i]))
                    else:
                        coupon_flow.append(coupon_flow[x - 1] + (coupon_flow[x - 1] * coupon / calculation_basis[i] * coupon_days[i]))
                else:
                    coupon_flow.append(coupon_flow[x - 1] * coupon / calculation_basis[i] * coupon_days[i])
            x += 1
            i += 1
        return coupon_flow

    def presentValue(self, current_coupon, nper, w_coupon, discrepancy, coupon_type, maturity_date, expected_maturity, coupon_days, ytm, periodicity, 
                     coupon_flow, payment_type, issuer, notional_value, gained_coupon): 
        
        fraction, factor = 0, 0
        preliminar_present_value, final_present_value = 0, 0
        clean_price, dirty_price, purchase_value = 0, 0, 0
        present_values, present_values_factor, values_cvx = [], [], []
        i = current_coupon
        
        while i <= nper:
            if (i + w_coupon - current_coupon) < 0:
                factor = 0
            else:
                if discrepancy == 'Yes' and i == nper:
                    if (coupon_type in ['fw1', 'fw2']):
                        fraction = 0
                    else:
                        fraction = ((maturity_date - expected_maturity).days) / coupon_days[nper - 1]
                factor = (i + w_coupon - current_coupon) + fraction
            
            # Calculate present value of each cash flow
            if issuer in ['Hacienda', 'hacienda'] and i == current_coupon:
                if current_coupon == nper:
                    present_values.append((coupon_flow[i - current_coupon] / (1 + (ytm / periodicity)) ** factor) -
                    (coupon_flow[i - current_coupon] * (1 - w_coupon)) + (notional_value / (1 + (ytm / periodicity)) ** factor))
                else:
                    present_values.append((coupon_flow[i - current_coupon] / (1 + (ytm / periodicity)) ** factor) -
                    (coupon_flow[i - current_coupon] * (1 - w_coupon)))
            else:
                present_values.append(coupon_flow[i - current_coupon] / (1 + ytm / periodicity) ** factor)
            
            # Calculate present value factors for duration
            present_values_factor.append(present_values[i - current_coupon] * factor)
            
            # Calculate values for convexity
            values_cvx.append(present_values[i - current_coupon] * (factor ** 2 + factor))
            i += 1
            
        # Conversion to numpy arrays for efficient calculations
        present_values = np.array(present_values)
        present_values_factor = np.array(present_values_factor) #necessary for the duration
        values_cvx = np.array(values_cvx) #necessary for convexity
        
        #Depending on the payment type, changes the present value of the security
        if (payment_type in ['pik']):
            preliminar_present_value = present_values [-1]
        else:
            preliminar_present_value = present_values.sum()
            
        # Determine final present value, clean price, and dirty price based on issuer
        if issuer not in ['Hacienda', 'hacienda']:
            final_present_value = preliminar_present_value
            dirty_price = round(final_present_value / notional_value,10)
            clean_price = round((final_present_value - gained_coupon) / notional_value,10)
            purchase_value = round(final_present_value - gained_coupon, 2)
            final_present_value = round(final_present_value,2)
        else:
            clean_price = round(preliminar_present_value / notional_value,6)
            final_present_value = round((notional_value * clean_price) + gained_coupon, 2)
            dirty_price = round(final_present_value / notional_value,6)
            purchase_value = round(notional_value * clean_price,2)
        
        # Calculate duration
        duration = (np.sum(present_values_factor) / preliminar_present_value) / periodicity
        
        # Calculate modified duration
        modified_duration = duration / (1 + (ytm / periodicity))
        
        # Calculate DV01
        dv01 = dirty_price * (modified_duration / 100) * notional_value / 100
        
        # Calculate convexity
        convexity = np.sum(values_cvx) / (1 + (ytm / periodicity)) ** 2 / preliminar_present_value / periodicity ** 2
        
        return final_present_value, dirty_price, purchase_value, clean_price, duration, modified_duration, dv01, convexity, preliminar_present_value, present_values

    def Repos(self, margin, repo_rate, repo_tenure):
        count = -1
        flow_dates = {}
        settlement_date = self.settlement_date
        gained_coupon, one_way_value, return_value, error = 0, 0, 0, 0
        
        try:
            if self.multiple: present_value = self.results['Preliminar Present Value' if self._lang == 'eng' else 'Valor Presente Preliminar']
            else: present_value = self.results['preliminar_present_value' if self._lang == 'eng' else 'valor_presente_preliminar']
        except Exception:
            error += 1
                
        if error > 0: return None, None
        
        coupon_dates = self.coupon_dates
        coupon_flows = self.coupon_flow
        repo_maturity = settlement_date + relativedelta(days =+ repo_tenure)
        if repo_maturity >= self.maturity_date:
            prompt = 'ERROR: Security matures before the repo..'
            return 'Maturing', 'Maturing'

        for date in coupon_dates:
            if date > settlement_date:
                count += 1
                flow_dates[date] = coupon_flows[count]

        for key, value in flow_dates.items():
            if key == repo_maturity: flow_dates[key] = value - self.notional_value
        
        for date in coupon_dates:
            if date > settlement_date and date < repo_maturity: gained_coupon += flow_dates[date]

        adjusted_present_value = present_value - gained_coupon
        one_way_value = adjusted_present_value * ( 1 - margin)
        return_value = one_way_value * (1 + repo_rate / 365 * repo_tenure)

        return one_way_value, return_value

    def FLR(self, rules: dict):
        asked_amount = self.face_value #this as a rule makes more sense
        available_amount, objective_notional = 0, 0
        if rules:
            rule = rules['gov'] if self.issuer.lower() in ['bcrd', 'hacienda'] else rules['other']
            rule = float(rule)  
        
        if self.multiple:
            present_value = self.results['Valor Presente Preliminar' if self._lang == 'esp' else 'Preliminar Present Value']
        else:
            present_value = self.results['valor_presente_preliminar' if self._lang == 'esp' else 'preliminar_present_value']

        try:
            dirty_price = present_value / self.notional_value
        except Exception as e:
            self._logger.error(f'Error calculating dirty price => {e}')
        available_amount = present_value / rule
        
        try:
            objective_notional = Support.solver(self.notional_value, dirty_price, asked_amount, rule) #se pudiera hacer el round a -4.. ponderar
        except Exception as e:
            self._logger.error(f'Error in the optimization => {e}')
            
        return available_amount, objective_notional
      
    def create_results(self, valuation_results, status):
        results = {}
        if self.multiple == False:
            results['valor_presente' if self._lang == 'esp' else 'present_value'] = valuation_results[0] if status else None
            results['precio' if self._lang == 'esp' else 'price'] = valuation_results[1] if status else None
            results['valor_compra' if self._lang == 'esp' else 'purchase_value'] = valuation_results[2] if status else None
            results['precio_limpio' if self._lang == 'esp' else 'clean_price'] = valuation_results[3] if status else None
            results['duracion' if self._lang == 'esp' else 'duration'] = valuation_results[4] if status else None
            results['duracion_modificada' if self._lang == 'esp' else 'modified_duration'] = valuation_results[5] if status else None
            results['dv01'] = valuation_results[6] if status else None
            results['convexidad' if self._lang == 'esp' else 'convexity'] = valuation_results[7] if status else None
            results['valor_presente_preliminar' if self._lang == 'esp' else 'preliminar_present_value'] = valuation_results[8] if status else None
            results['tiempo_calculo' if self._lang == 'esp' else 'process_time'] = self._process_time if status else None
        elif self.multiple == True:
            results['Emisor' if self._lang == 'esp' else 'Issuer'] = self.issuer
            results['Fecha Vencimiento' if self._lang == 'esp' else 'Maturity Date'] = self.maturity_date
            results['Nocional' if self._lang == 'esp' else 'Notional'] = self.notional_value
            results['Rendimiento' if self._lang == 'esp' else 'Yield'] = self.ytm
            results['Cupón Corrido' if self._lang == 'esp' else 'Running Coupon'] = self.gained_coupon if status else None
            results['Valor Presente' if self._lang == 'esp' else 'Present Value'] = valuation_results[0] if status else None
            results['Precio Sucio' if self._lang == 'esp' else 'Dirty Price'] = valuation_results[1] if status else None
            results['Valor Compra' if self._lang == 'esp' else 'Purchase Value'] = valuation_results[2] if status else None
            results['Precio Limpio' if self._lang == 'esp' else 'Clean Price'] = valuation_results[3] if status else None
            results['Duracion' if self._lang == 'esp' else 'Duration'] = valuation_results[4] if status else None
            results['Duracion Modificada' if self._lang == 'esp' else 'Modified Duration'] = valuation_results[5]  if status else None
            results['DV01'] = valuation_results[6] if status else None
            results['Convexidad' if self._lang == 'esp' else 'Convexity'] = valuation_results[7] if status else None
            results['Valor Presente Preliminar' if self._lang == 'esp' else 'Preliminar Present Value'] = valuation_results[8] if status else None
            results['Tiempo de Cálculo' if self._lang == 'esp' else 'Process Time'] = self._process_time if status else None
        return results
     
    def get_results(self):
            
        return self.results

    def __str__(self):
        try:
            if self._lang == 'eng':
                if self.settlement_date <= self.maturity_date:
                    return(f'\nSecurity Summary:\n \n \
Issuance Date: {self.issuance_date}\n \
Maturity Date: {self.maturity_date}\n \
Coupon Rate: {self.coupon}\n \
Yield to Maturity: {self.ytm}\n \
Payment Type: {self.payment_type}\n \
Coupon Type: {self.coupon_type}\n \
Periodicity: {self.periodicity}\n \
Methodology (basis): {self.methodology}\n \
Face Value: {self.face_value}\n \
Present Value: {self.results['present_value']}\n \
Price: {self.results['price']}\n \
Purchase Value: {self.results['purchase_value']}\n \
Clean Price: {self.results['clean_price']}\n \
Duration: {self.results['duration']}\n \
Modified Duration: {self.results['modified_duration']}\n \
DV01: {self.results['dv01']}\n \
Convexity: {self.results['convexity']}\n \
Preliminar Present Value: {self.results['preliminar_present_value']}\n \
Process Time: {self.results['process_time']}\n')
                else:
                    return f'Security matured by: {self.maturity_date - self.settlement_date} days'
            elif self._lang == 'esp':
                if self.settlement_date <= self.maturity_date:
                    return(f'\nResumen del Título:\n \n \
Fecha de Emisión: {self.issuance_date}\n \
Fecha de Vencimiento: {self.maturity_date}\n \
Tasa Cupón: {self.coupon}\n \
Rendimiento: {self.ytm}\n \
Tipo de Pago: {self.payment_type}\n \
Tipo de Cupón: {self.coupon_type}\n \
Periodicidad: {self.periodicity}\n \
Metodología (basis): {self.methodology}\n \
Monto: {self.face_value}\n \
Valor Presente: {self.results['valor_presente']}\n \
Precio: {self.results['precio']}\n \
Valor de Compra: {self.results['valor_compra']}\n \
Precio Limpio: {self.results['precio_limpio']}\n \
Duración: {self.results['duracion']}\n \
Duración Modificada: {self.results['duracion_modificada']}\n \
DV01: {self.results['dv01']}\n \
Convexidad: {self.results['convexidad']}\n \
Valor Presente Preliminar: {self.results['valor_presente_preliminar']}\n \
Tiempo de Cálculo: {self.results['tiempo_calculo']}\n')
                else:
                    return f'Título vencido por: {self.maturity_date - self.settlement_date} días'
        except Exception:
            return f'Security error!'

class Letras:
    def __init__(self, settlement_date, maturity_date, ytm, face_value, repo_margin = '', repo_rate = '', repo_tenure = '', date_format = '', multiple = False) -> None:
        self.settlement_date = settlement_date
        self.maturity_date = maturity_date
        self.ytm = ytm
        self.face_value = face_value
        self.repo_margin = repo_margin
        self.repo_rate = repo_rate
        self.repo_tenure = repo_tenure
        self.date_format = date_format
        self.multiple = multiple

        self.days_to_maturity = self.maturity_date - self.settlement_date
        self.price = 360 / (360 + self.ytm * self.days_to_maturity)
        self.purchase_value = self.face_value * self.price
        self.discount = self.face_value - self.purchase_value

class FiBackup:
    pass

    '''def amortizationDateValidation(self, coupon_dates, amortization_start_date, maturity_date, amortization_periods, amortization_periodicity):
        
        if (amortization_start_date in coupon_dates) == False:
            oldDate = amortization_start_date
            amortization_start_date = Support.closestDate(coupon_dates, amortization_start_date)
            prompt = '\n*ERROR! \nLa fecha de inicio de amortizaciones indicada no existe en los flujos de cupones, el programa procederá a buscar el cupon vigente más cercano.. ' + \
            '\n-Fecha Original: ' + str(oldDate.day) + '/' + str(oldDate.month) + '/' + str(oldDate.year) + \
            '\n-Fecha Nueva: ' + str(amortization_start_date.day) + '/' + str(amortization_start_date.month) + '/' + str(amortization_start_date.year)
            print(prompt , file = self.log)
            messagebox.showwarning('ValueError', prompt)
        fechaTest = amortization_start_date + relativedelta(months =+ ((amortization_periods - 1) * amortization_periodicity))
        if fechaTest > maturity_date:
            prompt = '\n*ERROR! \nLa fecha estimada de última amortización supera la fecha de vencimiento!' + \
                 '\n-Fecha de Vencimiento: ' + str(maturity_date.day) + '/' + str(maturity_date.month) + '/' + str(maturity_date.year) + \
                 '\n-Fecha estimada: ' + str(fechaTest.day) + '/' + str(fechaTest.month) + '/' + str(fechaTest.year)
            print(prompt, file = self.log)
            messagebox.showerror('ValueError', prompt)
            return amortization_start_date, False
        else: return amortization_start_date, True'''
        
    '''#This module executes all calculations related to amortization including validations (should be included more validations and user inputs through streamlit)
    def amortizationModule(self, payment_type, amortization_start_date, maturity_date, amortization_periods, amortization_periodicity, amortizable_percentage, 
                           settlement_date, face_value, coupon_dates, current_coupon, nper, periodicity):
        #First this runs validations
        while True:
            amortization_start_date, validation = self.amortizationDateValidation(coupon_dates, amortization_start_date, maturity_date, amortization_periods, amortization_periodicity)
            if validation == False:
                while True:
                    seleccion = input('\nQue desea modificar para solucionar el error? ' + 
                              '\n1-Fecha de Inicio de las Amortizaciones' + 
                              '\n2-Periodicidad de las Amortizaciones' + 
                              '\n3-Cantidad de Amortizaciones' + 
                              '\nRespuesta: ')
                    if seleccion in ['1','2','3']:
                        if seleccion == '1': 
                            amortization_start_date = self.inputFecha('Fecha de Inicio de Amotizaciones')
                            break
                        elif seleccion == '2':
                            while True:
                                amortization_periodicity = Support.periodicityConversion(self.inputPrompt('periodicidad de amortizaciones', ['mensual','m','1','bimensual','b','2','trimestral','t','3','cuatrimestral','c','4','semestral','s','5','anual','a','6'], 3), mode = 'amortization')
                                if (amortization_periodicity == 12 / periodicity) or (amortization_periodicity == (12 / periodicity) * 2):
                                    break
                                else:
                                    prompt = 'Valor inválido: la periodicidad de las amortizaciones debe de ser igual o el doble de la periodicidad de pagos de cupones..'
                                    print(prompt, file = self.log)
                                    messagebox.showerror('ValueError', prompt)
                                    amortization_periodicity, 12 / periodicity
                                    continue
                            break
                        elif seleccion == '3':
                            amortization_periods = self.inputIntegers()
                            break
                    else:
                        prompt = 'Opción inválida, inténtelo de nuevo..'
                        print(prompt, file = self.log)
                        messagebox.showerror('InputError', prompt)
                        continue
                continue
            else: break

        #Determination of the dates in which the securities amortize
        amortization_amount = 0
        amortization_dates = []
        dates_results = []
        if (payment_type in ['amortized', 'amortizado', 'pik']):
            date_count = 0
            if amortization_start_date < settlement_date:
                date_count = 1
            i = 0
            while i <= int(amortization_periods) - 1:
                if i == 0:
                    amortization_dates.append(amortization_start_date)
                else:
                    amortization_dates.append(amortization_dates[i - 1] + relativedelta(months=+amortization_periodicity))
                    if amortization_dates[i] <= settlement_date:
                        date_count += 1
                i += 1
            amortization_dates = np.array(amortization_dates)
            i = 0
            posicion = 0
            while i <= np.count_nonzero(amortization_dates) - 1:
                posicion = np.where(coupon_dates == amortization_dates[i])[0]
                if posicion != 0: dates_results.append(posicion)
                i += 1
            amortization_amount = (face_value * amortizable_percentage) / amortization_periods
            dates_results = np.array(dates_results)
        #Calculating the Notional Value and the Adjusted Amounts
        previously_amortized = 0
        notional_value = 0
        adjusted_value = []
        if (payment_type in ['amortized', 'amortizado']):
            if amortization_periods > 0:
                if np.count_nonzero(dates_results) != date_count:
                    previously_amortized = amortization_amount * date_count
                notional_value = face_value - previously_amortized
            count = 0
            i = current_coupon
            while i <= nper:
                if np.where(dates_results == i)[0].size > 0:
                    count += 1
                    if count <= (amortization_periods - (amortization_periods - np.count_nonzero(dates_results))):
                        adjusted_value.append(notional_value - (amortization_amount * (count - 1)))
                    else:
                        adjusted_value.append(face_value)
                else:
                    adjusted_value.append(notional_value - (amortization_amount * count))
                i += 1
            adjusted_value = np.array(adjusted_value)

        return notional_value, amortization_amount, adjusted_value, dates_results'''


#from pathlib import Path
#import sys, os
#sys.path.append(str(Path(__file__).resolve().parent.parent))
import numpy as np, pandas as pd, matplotlib.pyplot as plt, logging
from caluxPy_fi.FixedIncome import FixedIncome
from caluxPy_fi.Support import Support

class Convexity(FixedIncome):

    def __init__(self, settlement_date, issuance_date, maturity_date, coupon, ytm, face_value, issuer, methodology, periodicity, payment_type, coupon_type, 
                 forward_date, amortization_start_date, amortization_periods, amortization_periodicity, amortizable_percentage, date_format = '', multiple = False, bps = 100, lang = 'eng'):
        super().__init__(settlement_date, issuance_date, maturity_date, coupon, ytm, face_value, issuer, methodology, periodicity, payment_type, coupon_type, 
                         forward_date, amortization_start_date, amortization_periods, amortization_periodicity, amortizable_percentage, date_format = '', multiple = False, lang = 'eng')

        desktop = Support.get_folder_path()
        logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler("{0}/{1}.log".format(desktop, 'convex'), mode = 'w', encoding = 'utf-8')
        fileHandler.setFormatter(logFormatter)
        self._logger.addHandler(fileHandler)

        self._lang = lang
        self.tipo_valoracion = 'convexity'
        self.bps = bps

        self.max_gain, self.max_loss, self.avg_gain, self.avg_loss, self.std_gain, self.std_loss = 0, 0, 0, 0, 0, 0
        self.chart_data = {}
        
        try:
            self.convexityValuationType(self.current_coupon, self._nper, self.w_coupon, self._discrepancy, self.coupon_type, self.maturity_date, self.expected_maturity, self.coupon_days, 
                                        self.ytm, self.periodicity, self.coupon_flow, self.payment_type, self.issuer, self.notional_value, self.gained_coupon, bps)
            self._logger.info('SUCCESS: Calculation finished..')
        except Exception as e:
            self._logger.exception(e)

        self._logger.removeHandler(fileHandler)
        fileHandler.close()

    def convexityValuationType(self, current_coupon, nper, w_coupon, discrepancy, coupon_type, maturity_date, expected_maturity, coupon_days, ytm, periodicity, 
                                 coupon_flow, payment_type, issuer, notional_value, gained_coupon, bps):
        
        listPBS, listDirtyPrice, listConvexity, listDuration = [], [], [], []

        y = -int(bps)
        while y <= int(bps):
            ytm2 = ytm + (y / 10000)
            resValuation = self.presentValue(current_coupon = current_coupon, 
                                             nper = nper, 
                                             w_coupon = w_coupon, 
                                             discrepancy = discrepancy, 
                                             coupon_type = coupon_type, 
                                             maturity_date = maturity_date, 
                                             expected_maturity = expected_maturity, 
                                             coupon_days = coupon_days, 
                                             ytm = ytm2, 
                                             periodicity = periodicity, 
                                             coupon_flow = coupon_flow, 
                                             payment_type = payment_type, 
                                             issuer = issuer, 
                                             notional_value = notional_value, 
                                             gained_coupon = gained_coupon)
            # Append results to lists
            listPBS.append(y) # List of basis points
            listDirtyPrice.append(resValuation[1] * 100) # List of dirty prices
            listDuration.append(resValuation[4]) # if y != 0 else 0)
            listConvexity.append(resValuation[7]) # if y != 0 else 0)
            y += 1 # Marginal increase
        
        # Conversion of lists into numpy arrays for ease of use
        listPBS = np.array(listPBS)
        listDirtyPrice = np.array(listDirtyPrice)
        listDuration = np.array(listDuration)
        listConvexity = np.array(listConvexity)
        
        # Creation of DataFrame
        self.df_convex = pd.DataFrame({
            'pbs': listPBS, 
            'Price': listDirtyPrice,
            'Duration': listDuration,
            'Convexity': listConvexity
        })
        
        # Process negative PBS values
        negatives = self.df_convex[self.df_convex['pbs'] <= 0].copy() # Copy the DataFrame for ease of use
        negatives.sort_values(by = 'pbs', ascending = False, inplace = True) # Sort descending
        negatives['dv01'] = negatives['Price'].diff().fillna(0) * 10000 # Calculate the difference and fill first with 0
        negatives['gains/losses'] = negatives['dv01'].cumsum() # Perform Cummulative Sum
        #negatives[['Duration', 'Convexity']] = negatives[['Duration', 'Convexity']].cumsum()
        negatives.sort_values(by = 'pbs', ascending = True, inplace = True) # Sort Ascending
        first_neg = negatives.iloc[-2]['dv01'] # Get the first negative (the one after 0)
        
        # Treatment of the positives in the df
        positives = self.df_convex[self.df_convex['pbs'] >= 0].copy() # Copy the DataFrame for ease of use
        positives.sort_values(by = 'pbs', ascending = True, inplace = True) # Sort Ascending
        positives['dv01'] = positives['Price'].diff().fillna(0) * -10000 # Calculate the difference and fill first with 0
        positives['gains/losses'] = positives['dv01'].cumsum() # Perform Cummulative Sum
        #positives[['Duration', 'Convexity']] = positives[['Duration', 'Convexity']].cumsum()
        first_pos = positives.iloc[1]['dv01'] # Get the first positive (the one after 0)
        
        # Calculate average dv01
        self.avg_dv01 = (first_neg + first_pos) / 2 # Get an averaged dv01 with the first occurrence of a marginal increase, because it can go both ways
        
        #Concatenating the dfs to form the complete one
        final = pd.concat([negatives, positives[1:]]) # Concatenate both results of gains and losses
        final.reset_index(drop = True, inplace = True) # Reset indexes so it can be incorporated into the original DataFrame
        self.df_convex[['dv01', 'gains/losses', 'Duration', 'Convexity']] = final[['dv01', 'gains/losses', 'Duration', 'Convexity']] # Incorporate gains and losses into the original DataFrame
        
        #Proceed over a copy so the original is not affected
        final_copy = final.copy()
        final_copy = final_copy[final_copy['pbs'] != 0] # Exclude 0 bps row, so no contamination of calculations
        final_copy['Negative'] = final_copy['pbs'] < 0 # Add a boolean column where True is if value of pbs is negative
        max_values_by_category = final_copy.groupby('Negative')['dv01'].max() # Group the data by the boolean column, having 2 groups (True and False) then getting max values of box, resulting into a list
        avg_values_by_category = final_copy.groupby('Negative')['dv01'].mean() # Get a list with the average means of both groups
        std_values_by_category = final_copy.groupby('Negative')['dv01'].std() # Get a list with the standard deviations of both groups

        # Assigning values to the usable variables
        self.max_gain = max_values_by_category[True] 
        self.max_loss = max_values_by_category[False]
        self.avg_gain = avg_values_by_category[True]
        self.avg_loss = avg_values_by_category[False]
        self.std_gain = std_values_by_category[True]
        self.std_loss = std_values_by_category[False]
       
    def showGraph(self):
        plt.figure(figsize=(10,6))
        plt.plot(self.chart_data['x_pbs'], self.chart_data['y1_dur'])
        plt.plot(self.chart_data['x_pbs'], self.chart_data['y2_cvx'], linestyle='--')
        title = 'Gráfico de Duración-Convexidad' if self._lang == 'esp' else 'Duration-Convexity Graph' if self._lang == 'eng' else ''
        basis_points = 'Puntos Básicos' if self._lang == 'esp' else 'Basis Points' if self._lang == 'eng' else ''
        values = 'Valores Duración y Convexidad' if self._lang == 'esp' else 'Duration and Convexity Values' if self._lang == 'eng' else ''
        plt.ylabel(values)
        plt.xlabel(basis_points)
        plt.title(title)
        
        return plt.show()

    def saveGraph(self, filename):
        plt.savefig(filename, bbox_inches = 'tight', dpi = 300)
        plt.close()
    
    def getResults(self):
        results = {'max_gain': self.max_gain,
                   'max_loss': self.max_loss, 
                   'avg_gain': self.avg_gain, 
                   'avg_loss': self.avg_loss, 
                   'sd_gain': self.std_gain, 
                   'sd_loss': self.std_loss,
                   'dv01': self.avg_dv01,
                   'df': self.df_convex,
                   #'chart_data': self.chart_data
                   }
        return results

    def __str__(self):
        return f'\
{'Max Gain' if self._lang == 'eng' else 'Max Ganancia'}: {self.max_gain} \n \
{'Max Loss' if self._lang == 'eng' else 'Max Perdida'}: {self.max_loss} \n \
{'Average Gain' if self._lang == 'eng' else 'Ganancia Promedio'}: {self.avg_gain} \n \
{'Average Loss' if self._lang == 'eng' else 'Perdida Promedio'}: {self.avg_loss} \n \
{'Deviations in Gains' if self._lang == 'eng' else 'Desviacion Ganancias'}: {self.std_gain} \n \
{'Deviations in Losses' if self._lang == 'eng' else 'Desviacion Perdidas'}: {self.std_loss}'






#import sys, os
#from pathlib import Path
#sys.path.append(str(Path(__file__).resolve().parent.parent))
import pandas as pd, time, logging
from caluxPy_fi.FixedIncome import FixedIncome
from caluxPy_fi.Support import Support

class Multiple:

    def __init__(self, data: pd.DataFrame, additional_ops: dict = {}, lang = 'eng'):

        desktop = Support.get_folder_path()
        logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        _logger = logging.getLogger()
        _logger.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler("{0}/{1}.log".format(desktop, 'cMult'), mode = 'w', encoding = 'utf-8')
        fileHandler.setFormatter(logFormatter)
        _logger.addHandler(fileHandler)
    
        self.start_time = time.perf_counter()
        self.data = data.to_dict('records')
        self.additional_ops = additional_ops
        self.mResults = []
        i = 0
        while i < len(self.data):
            #Normal Attributes
            for key, values in self.data[i].items():
                setattr(self, key[2:], values)
            #Additional Attributes passed through the class
            if len(additional_ops) >= 1:
                for op, attributes in additional_ops.items():
                    for attribute, value in attributes.items():
                        setattr(self, attribute, value)
            
            try:
                repos_margin = self.margin if lang == 'eng' else self.margen if lang == 'esp' else ''
            except Exception:
                repos_margin = None
            try:
                repo_rate = self.interestRate if lang == 'eng' else self.tasaInteres if lang == 'esp' else ''
            except Exception:
                repo_rate = None
            try:
                repo_tenure = self.repoTenure if lang == 'eng' else self.plazoRepos if lang == 'esp' else ''
            except Exception:
                repo_tenure = None

            try:
                rules = self.rules if lang == 'eng' else self.reglas if lang == 'esp' else ''
            except Exception:
                rules = {}

            try:
                ratio = self.ratio
            except Exception as e:
                ratio = ''
            
            settlement_date = self.settlementDate if lang == 'eng' else self.fechaLiquidacion if lang == 'esp' else ''
            issuance_date = self.issuanceDate if lang == 'eng' else self.fechaEmision if lang == 'esp' else ''
            maturity_date = self.maturityDate if lang == 'eng' else self.fechaVencimiento if lang == 'esp' else ''
            coupon = self.coupon if lang == 'eng' else self.cupon if lang == 'esp' else ''
            ytm = self.ytm if lang == 'eng' else self.rendimiento if lang == 'esp' else ''
            face_value = self.facialValue if lang == 'eng' else self.monto if lang == 'esp' else ''
            issuer = self.issuer if lang == 'eng' else self.emisor if lang == 'esp' else ''
            methodology = self.methodology if lang == 'eng' else self.metodologia if lang == 'esp' else ''
            periodicity = self.periodicity if lang == 'eng' else self.periodicidad if lang == 'esp' else ''
            payment_type = self.paymentType if lang == 'eng' else self.tipoPago if lang == 'esp' else ''
            coupon_type = self.couponType if lang == 'eng' else self.tipoCupones if lang == 'esp' else ''
            forward_date = self.forwardDate if lang == 'eng' else self.forwardDate if lang == 'esp' else ''
            amortization_start_date = self.amotizationDate if lang == 'eng' else self.fechaAmortizacion if lang == 'esp' else ''
            amortization_periods = self.amortizationTimes if lang == 'eng' else self.cantidadAmortizaciones if lang == 'esp' else ''
            amortization_periodicity = self.amortizationPeriodicity if lang == 'eng' else self.periodicidadAmortizaciones if lang == 'esp' else ''
            amortizable_percentage = self.amortizablePercentage if lang == 'eng' else self.porcentajeAmortizable if lang == 'esp' else ''
            
            fi_instance = FixedIncome(settlement_date = settlement_date, 
                                      issuance_date = issuance_date, 
                                      maturity_date = maturity_date, 
                                      coupon = coupon, 
                                      ytm = ytm, 
                                      face_value = face_value, 
                                      issuer = issuer, 
                                      methodology = methodology, 
                                      periodicity = periodicity, 
                                      payment_type = payment_type, 
                                      coupon_type = coupon_type, 
                                      forward_date = forward_date, 
                                      amortization_start_date = amortization_start_date, 
                                      amortization_periods = amortization_periods, 
                                      amortization_periodicity = amortization_periodicity, 
                                      amortizable_percentage = amortizable_percentage, 
                                      date_format = '%Y-%m-%d', 
                                      multiple = True,
                                      repo_margin = repos_margin,
                                      repo_rate = repo_rate,
                                      repo_tenure = repo_tenure,
                                      flr_rules = rules,
                                      flr_ratio = ratio,
                                      lang = lang)
            calculador_results = fi_instance.get_results()
            self.mResults.append(calculador_results)
            prompt = f'Calculated - {i + 1} of {len(data)}' if lang == 'eng' else f'Calculados - {i + 1} de {len(data)}'
            _logger.info(f'{prompt} : ({issuer} | {maturity_date})')
            i += 1

        self.df_resultados_multiples = pd.DataFrame(self.mResults)
        self.end_time = time.perf_counter()
        self.process_time = self.end_time - self.start_time
        _logger.info(f'{self.process_time} {'secs' if lang == 'eng' else 'segs'}')
        _logger.removeHandler(fileHandler)
        fileHandler.close()

    def get_results(self):

        return self.mResults

    def get_results_df(self):

        return self.df_resultados_multiples
    
    def processTime(self):

        return self.process_time


import openpyxl, pandas as pd, numpy as np, datetime, sys, time, os, logging
from collections import Counter
from caluxPy_fi.Support import Support

class ExternalImport:
    
    def __init__(self, file = '', df = '', file_type = '', initial_row = 0, columns = '', column_type = {}):#, repos = False, repos_attributes = [], medida = False):
        self.file = file
        self.file_type = file_type
        self.initial_row = initial_row
        self.columns = columns
        self.df = None
        self.setup_logging()
        self.read_file()
        self.validate_data()

    def setup_logging(self):
        desktop = Support.get_folder_path()
        logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler(f"{desktop}/ei.log", mode='w', encoding='utf-8')
        fileHandler.setFormatter(logFormatter)
        self.logger.addHandler(fileHandler)

    def read_file(self):
        if self.file:
            try:
                if self.file_type == '.csv':
                    self.df = pd.read_csv(self.file, header=self.initial_row)
                elif self.file_type == '.xlsx':
                    self.df = pd.read_excel(self.file, header=self.initial_row)
                else:
                    self.logger.error(f'Unsupported file type: {self.file_type}')
                    raise ValueError(f'Unsupported file type: {self.file_type}')
            except Exception as e:
                self.logger.exception(f'Failed to read file: {self.file}. Error: {e}')
                raise

    def validate_data(self):
        # Splitting validation into different aspects
        self.validate_dates()
        self.validate_numbers()
        self.validate_percentages()
        self.cleanup_errors()
        
    def validate_dates(self):
        date_columns = [col for col in self.df.columns if col.startswith('d')]
        for col in date_columns:
            self.df[col] = pd.to_datetime(self.df[col], errors='coerce')
            # Further custom validation can be implemented here

    def validate_numbers(self):
        number_columns = [col for col in self.df.columns if col.startswith('n')]
        for col in number_columns:
            self.df[col] = pd.to_numeric(self.df[col], errors='coerce')
            # Add any specific number validation here

    def validate_percentages(self):
        percentage_columns = [col for col in self.df.columns if col.startswith('p')]
        for col in percentage_columns:
            self.df[col] = self.df[col].apply(self.convert_percentage)
            # Add any specific percentage validation here

    def convert_percentage(self, value):
        try:
            if isinstance(value, str) and '%' in value:
                return float(value.strip('%')) / 100
            return float(value)
        except ValueError:
            self.errors.append(f"Invalid percentage value: {value}")
            return pd.NA
    
    def cleanup_errors(self):
        # Example of removing rows based on certain validation criteria
        self.df.dropna(inplace=True)  # This is a placeholder for actual error handling logic

    def get_results(self):
        return self.df.to_dict('records')   
    
    def __del__(self):
        # Ensure the logging handlers are properly closed
        handlers = self.logger.handlers[:]
        for handler in handlers:
            handler.close()
            self.logger.removeHandler(handler)         
        
    def getResults(self):

        return self.importedData

    def processTime(self):

        return self.tiempo_calculo

    def conteoErrores(self, my_dict):
        conteo = 0
        # Create a Counter object for each list in the dictionary
        counters = {key: Counter(values) for key, values in my_dict.items()}
        # Display the counts for each list in the dictionary
        for key, counter in counters.items():
            #print(f"Conteo para '{key}':")
            for element, count in counter.items():
            #    print(f"{element}: {count}")
                conteo += count
        return conteo


import pandas as pd, numpy as np, time, math, logging, os
from scipy.optimize import minimize
from caluxPy_fi.Support import Support

class MPT:
    
    def __init__(self, data, yearBasis = None, numberAssets = None, maxWeight = None, riskAversion = None):
        
        desktop = Support.get_folder_path()
        logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler("{0}/{1}.log".format(desktop, 'mpt'), mode = 'w', encoding = 'utf-8')
        fileHandler.setFormatter(logFormatter)
        logger.addHandler(fileHandler)
        
        self.assetPrices = data
        self.numberAssets = numberAssets
        self.maxWeight = maxWeight
        if self.maxWeight is None or self.maxWeight == '': self.maxWeight = 1.0
        self.riskAversion = riskAversion
        if self.riskAversion is None or self.riskAversion == '': self.riskAversion = 1
        self.yearBasis = int(yearBasis)
        if self.yearBasis is None or self.yearBasis == '': self.yearBasis = 365
        self.totalTime = 0
        
        if self.numberAssets is None or self.numberAssets == '': self.numberAssets = len(self.assetPrices)

        self.stocks = list(self.assetPrices.columns.values)

        logger.info('ATTENTION: Calculating yields of assets..')
        self.assetYields, self.processingTime = self.calculate_yields(self.assetPrices) #not ln
        logger.info(f'ATTENTION: Yield of assets calculated.. - Time Elapsed: {self.processingTime} s \n')
        self.totalTime += self.processingTime

        self.averageYields = self.effectiveRate(self.assetYields.mean(),self.yearBasis)
        self.devYields = self.assetYields.std() * math.sqrt(self.yearBasis)
        self.sharpeRatios = self.averageYields / self.devYields

        logger.info('ATTENTION: Creating Correlation Matrix..')
        self.correlationMatrix, self.processingTime = self.calculate_correlation_matrix(self.assetYields, self.numberAssets)
        logger.info(f'ATTENTION: Correlation Matrix Created.. - Time Elapsed: {self.processingTime} s \n')
        self.totalTime += self.processingTime

        #self.covarianceMatrix = self.calculate_varCovar_matrix(self.correlationMatrix, self.devYields, self.numberAssets, self.stocks)
        logger.info('ATTENTION: Creating Covariance Matrix..')
        self.covarianceMatrix, self.processingTime = self.calculate_varCovar_matrix_2(self.correlationMatrix, self.devYields, self.numberAssets) #faster
        logger.info(f'ATTENTION: Covariance Matrix Created.. - Time Elapsed: {self.processingTime} s \n')
        self.totalTime += self.processingTime

        logger.info('ATTENTION: Generating Results..')
        self.optimalPortfolio, self.processingTime = self.optimize_portfolio(self.averageYields, self.covarianceMatrix, self.numberAssets, self.maxWeight, self.stocks)
        
        logger.info(f'ATTENTION: Results Generated.. - Time Elapsed: {self.processingTime} s \n')
        self.totalTime += self.processingTime

        logger.info(f'Process Completed! Total Time: {self.totalTime} s')
        
        logger.removeHandler(fileHandler)
        fileHandler.close()
        
    def calculate_yields(self, stock_prices):
        tiempo_inicio_calculo = time.perf_counter()
        # Calculate the percentage change for all stock columns
        #yields_df = stock_prices.drop('Date', axis=1).pct_change()
        yields_df = stock_prices.pct_change()
        # If you prefer to see the yields as percentages, multiply by 100
        #yields_df *= 100
    
        # Add the 'Date' column back to the DataFrame
        #yields_df.insert(0, 'Date', stock_prices['Date'])
    
        # Drop the first row (it will have NaN values due to the percentage change calculation)
        yields_df = yields_df.dropna()
        tiempo_final_calculo = time.perf_counter()
        processing_time = tiempo_final_calculo - tiempo_inicio_calculo
        return yields_df, processing_time

    def calculate_correlation_matrix(self, stock_yields, num_assets):
        tiempo_inicio_calculo = time.perf_counter()
        # Calculate yields
        #stock_prices.set_index('Date', inplace=True)
    
        # Select the yields of the specified stocks
        selected_yields = stock_yields.iloc[:,:num_assets]
    
        # Calculate the correlation matrix
        correlation_matrix = selected_yields.corr()
        tiempo_final_calculo = time.perf_counter()
        processing_time = tiempo_final_calculo - tiempo_inicio_calculo
        return correlation_matrix, processing_time

    def calculate_covariance_matrix(self,stock_yields, num_assets):
        # Calculate yields
        #stock_prices.set_index('Date', inplace=True)
        #stock_yields = calculate_yields(stock_prices)
    
        # Select the yields of the specified stocks
        selected_yields = stock_yields.iloc[:,:num_assets]
    
        # Calculate the covariance matrix
        covariance_matrix = selected_yields.cov()
    
        return covariance_matrix

    def calculate_varCovar_matrix(self, correlation_matrix, standard_deviations, num_assets, stock_names):
        tiempo_inicio_calculo = time.perf_counter()
        varCovar_matrix = []
        for y in range(num_assets):
            row = []
            for x in range(num_assets):
                row.append(correlation_matrix.iloc[y,x] * standard_deviations[y] * standard_deviations[x])
            varCovar_matrix.append(row)
        varCovar_matrix = pd.DataFrame(varCovar_matrix, columns=stock_names[:num_assets])
        tiempo_final_calculo = time.perf_counter()
        print(f'Processing Time: {tiempo_final_calculo - tiempo_inicio_calculo}')
        return varCovar_matrix

    def calculate_varCovar_matrix_2(self, correlation_matrix, standard_deviations, num_assets):
        tiempo_inicio_calculo = time.perf_counter()
        selected_stocks = standard_deviations[:num_assets]
        diagonal_matrix = np.diag(selected_stocks)
        varCovar_matrix = diagonal_matrix @ correlation_matrix @ diagonal_matrix.T #@ is same as dot for matrix multiplication
        tiempo_final_calculo = time.perf_counter()
        processing_time = tiempo_final_calculo - tiempo_inicio_calculo
        return varCovar_matrix, processing_time

    def effectiveRate(self, yields, basis):
        effectiveYields = (1 + yields) ** basis - 1
        return effectiveYields

    def optimize_portfolio(self, returns, cov_matrix, num_assets, max_weight, stocks):
        tiempo_inicio_calculo = time.perf_counter()
        results = pd.DataFrame()
        #Sequential Least Squares Programming (SLSQP)
        stocks = stocks[:num_assets]
        # Risk-free rate (e.g., 10-year Treasury yield)
        risk_free_rate = 0.03

        # Total number of assets
        returns = returns[:num_assets] 

        # Constraint: The sum of weights must equal max_weight
        constraint = {'type': 'eq', 'fun': lambda x: np.sum(x) - 1}

        # Constraint: Weights should be between 0 and 1
        bound = (0.0, max_weight)
        bounds = [bound] * num_assets

        # Define the termination tolerance
        tolerance = 1e-10

        # Objective function to maximize (negative of the Sharpe Ratio)
        def neg_sharpe_ratio(weights):
            portfolio_return = np.sum(returns * weights)
            portfolio_stddev = np.sqrt(np.dot(weights.T, np.dot(cov_matrix, weights)))
            #sharpe_ratio = (portfolio_return - risk_free_rate) / portfolio_stddev
            sharpe_ratio = (portfolio_return) / portfolio_stddev
            return -sharpe_ratio
        
        def expected_return(weights):
            return -np.sum(returns * weights)

        def standard_deviation(weights):
            return np.sqrt(np.dot(weights.T, np.dot(cov_matrix, weights)))

        # Complete portfolio optimization function
        #def optimize_portfolio():
        initial_weights = np.ones(num_assets) / num_assets  # Initialize with equal weights

        #region Equal Weightings Region
        dfResult = []
        weights = initial_weights
        portfolio_return = np.sum(returns * weights)
        portfolio_stddev = np.sqrt(np.dot(weights.T, np.dot(cov_matrix, weights)))
        sharpe = portfolio_return / portfolio_stddev
        dfResult = pd.DataFrame(weights, index=stocks, columns=['Equal Weighting'])
        dfResult.loc['Return'] = portfolio_return
        dfResult.loc['StDev'] = portfolio_stddev
        dfResult.loc['Sharpe'] = sharpe
        dfResult.loc['Risk Aversion'] = self.riskAversion
        dfResult.loc['Investor Utility'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2)
        dfResult.loc['A01'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2) - (portfolio_return - (0.5* (self.riskAversion + 1) * portfolio_stddev ** 2))
        results = dfResult
        #endregion

        #region Maximum Sharpe Ratio Region
        dfResult = []
        result = minimize(neg_sharpe_ratio, initial_weights, method='SLSQP', bounds=bounds, constraints=constraint, options={'ftol': tolerance})
        if result.success:
            weights = np.array(result.x)
            portfolio_return = np.sum(returns * weights)
            portfolio_stddev = np.sqrt(np.dot(weights.T, np.dot(cov_matrix, weights)))
            sharpe = -result.fun  # Convert back to positive value
            dfResult = pd.DataFrame(weights, index=stocks, columns=['Max Sharpe'])
            dfResult.loc['Return'] = portfolio_return
            dfResult.loc['StDev'] = portfolio_stddev
            dfResult.loc['Sharpe'] = sharpe
            dfResult.loc['Risk Aversion'] = self.riskAversion
            dfResult.loc['Investor Utility'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2)
            dfResult.loc['A01'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2) - (portfolio_return - (0.5* (self.riskAversion + 1) * portfolio_stddev ** 2))
            results = pd.concat([results, dfResult], axis = 1)
        else:
            print('Unsuccessful!')
        #endregion

        #region Maximum Returns Region
        dfResult = []
        result = minimize(expected_return, initial_weights, method='SLSQP', bounds=bounds, constraints=constraint, options={'ftol': tolerance})
        if result.success:
            weights = np.array(result.x)
            portfolio_return = -result.fun  # Convert back to positive value
            portfolio_stddev = np.sqrt(np.dot(weights.T, np.dot(cov_matrix, weights)))
            sharpe = portfolio_return / portfolio_stddev
            dfResult = pd.DataFrame(weights, index=stocks, columns=['Max Return'])
            dfResult.loc['Return'] = portfolio_return
            dfResult.loc['StDev'] = portfolio_stddev
            dfResult.loc['Sharpe'] = sharpe
            dfResult.loc['Risk Aversion'] = self.riskAversion
            dfResult.loc['Investor Utility'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2)
            dfResult.loc['A01'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2) - (portfolio_return - (0.5* (self.riskAversion + 1) * portfolio_stddev ** 2))
            results = pd.concat([results, dfResult], axis = 1)
        else:
            print('Unsuccessful!')
        #endregion

        #Minimum Risk Region
        #region
        result = minimize(standard_deviation, initial_weights, method='SLSQP', bounds=bounds, constraints=constraint, options={'ftol': tolerance})
        if result.success:
            weights = np.array(result.x)
            portfolio_return = np.sum(returns * weights)
            portfolio_stddev = result.fun  # Convert back to positive value
            sharpe = portfolio_return / portfolio_stddev
            dfResult = pd.DataFrame(weights, index=stocks, columns=['Min Risk'])
            dfResult.loc['Return'] = portfolio_return
            dfResult.loc['StDev'] = portfolio_stddev
            dfResult.loc['Sharpe'] = sharpe
            dfResult.loc['Risk Aversion'] = self.riskAversion
            dfResult.loc['Investor Utility'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2)
            dfResult.loc['A01'] = portfolio_return - (0.5 * self.riskAversion * portfolio_stddev ** 2) - (portfolio_return - (0.5* (self.riskAversion + 1) * portfolio_stddev ** 2))
            results = pd.concat([results, dfResult], axis = 1)
        else:
            print('Unsuccessful!')
        #endregion
        tiempo_final_calculo = time.perf_counter()
        processing_time = tiempo_final_calculo - tiempo_inicio_calculo
        return results, processing_time

    def showImportedData(self):
        return self.assetPrices

    def showYieldData(self):
        return self.assetYields

    def showCorrelationMatrix(self):
        return self.correlationMatrix

    def showCovarianceMatrix(self):
        return self.covarianceMatrix

    def showOptimalSolution(self):
        return self.optimalPortfolio


#import openpyxl, numpy as np, os
import pandas as pd, datetime, sys, time, logging, numpy as np
from collections import Counter
from caluxPy_fi.Support import Support

class ExternalImport:
    
    def __init__(self, file = '', df = '', file_type = '', initial_row = 0, columns = ''):#, repos = False, repos_attributes = [], medida = False):
        self.file = file
        self.df = df
        self.file_type = file_type
        self.initial_row = initial_row
        self.columns = columns
        
        #self.repos = repos
        #self.repos_attributes = repos_attributes
        #self.medida = medida
        self.tiempo_inicio_calculo = time.perf_counter()

        desktop = Support.get_folder_path()

        logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler("{0}/{1}.log".format(desktop, 'ei'), mode = 'w', encoding = 'utf-8')
        fileHandler.setFormatter(logFormatter)
        logger.addHandler(fileHandler)

        if self.file != '':
            try:
                if self.file_type == '.csv':
                    self.df = pd.read_csv(self.file, index_col = None, header = self.initial_row)
                else:
                    self.df = pd.read_excel(self.file, index_col = None, header = self.initial_row) #index col makes column its index, none creates one
            except Exception:
                logger.exception(f'ERROR: Archivo - {file} - no existe! Saliendo..\n')
                logger.removeHandler(fileHandler)
                fileHandler.close()
                sys.exit()

        logger.info(f'AVISO: Se iniciará el procedimiento de validación de las informaciones - líneas: {len(self.df)} columnas: {len(self.df.columns)}\n')
        columna = list(self.df.columns.values)
        dicErrores = {}
        #Procedimiento de Validación
        for i in range(len(self.df.columns)):
            counter = 0
            errores = []
            #caso para las fechas:
            #1 - Se convierten todos los datos a formato datetime.date, si hay valores vacíos se pondrá NaT 
            #2 - Se itera por cada una de las celdas de la columna
            #3 - Si al convertir el valor es NaT se procede
            #4 - Valida si la columna es Fecha de liquidación fwd o inicio de amortizaciones
            #4.1 - Si es Fecha de Liquidación, se sustituye el NaT por la fecha del día de hoy
            #4.2 - Si es Fecha Fwd, se procede
            #4.2.1 - Se valida si el tipo de pago de cupón es fw3, debido a que este tipo necesita una fecha, los otros se calculan automáticos
            #4.2.1 - Si es fw3 se agrega a errores, si no es, se quita el NaT y se deja en blanco
            #4.3 - Si es fecha de inicio de amortización primero se verifica si el tipo de pago es bullet, y 
            #4.3.1 - Si es bullet se sustituye por blanco ya que bullet no tiene estas fechas, sino se lleva a errores
            #4.4 - Cualquier otra columna si tiene error, se lleva a errores

            if columna[i][0] == 'd': #--> Dates
                logger.info(f'Trabajando Columna {columna[i]}')
                self.df[columna[i]] = pd.to_datetime(self.df[columna[i]], format='%Y-%m-%d', errors='coerce').dt.date # 1
                for cell in self.df[columna[i]]: # 2
                    if str(cell) == 'NaT': # 3
                        if columna[i] in ['d_settlementDate', 'd_forwardDate', 'd_amotizationDate']: # 4
                            if columna[i] == 'd_settlementDate': # 4.1
                                self.df.iloc[counter,i] = pd.to_datetime(datetime.date.today())
                            elif columna[i] == 'd_forwardDate': # 4.2
                                if self.df['s_couponType'][counter] == 'fw3': # 4.2.1
                                    logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es una fecha - Registro se sacará de la data.\n')
                                    errores.append(counter)
                                else:
                                    self.df.iloc[counter, i] = pd.NaT
                            elif columna[i] == 'd_amotizationDate': # 4.3
                                if self.df['t_paymentType'][counter] == 'bullet': # 4.3.1
                                    self.df.iloc[counter, i] = pd.NaT
                                else: #4.4
                                    logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es una fecha - Registro se sacará de la data.\n')
                                    errores.append(counter) 
                        else:
                            logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es una fecha - Registro se sacará de la data.\n')
                            errores.append(counter)
                    counter += 1
            #caso para montos
            #1 - Primero se convierte todo a float, si hay errores se convierten a NaM
            elif columna[i][0] == 'n': #--> Numbers
                self.df[columna[i]] = pd.to_numeric(self.df[columna[i]], errors = 'coerce') # 1
                for cell in self.df[columna[i]]:
                    if str(cell) == 'nan':
                        if columna[i] == 'n_amortizationTimes':
                            if self.df['t_paymentType'][counter] == 'bullet':
                                self.df.iloc[counter, i] = np.nan
                            else:
                                logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es un valor aceptado - Registro se sacará de la data.\n')
                                errores.append(counter)
                        else:
                            logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es un valor aceptado - Registro se sacará de la data.\n')
                            errores.append(counter)
                    elif cell < 0:
                        logger.info(f'ValueError: -línea {counter}- en -{columna[i]}- es menor que 0 - Registro se convertirá a positivo.\n')
                        self.df.iloc[counter, i] = -cell
                    counter += 1
            #caso para tasas
            elif columna[i][0] == 'p': #--> Percentages
                dato = ''
                for cell in self.df[columna[i]]:
                    if cell == '':
                        if columna[i] == 'p_coupon' or columna[i] == 'p_ytm':
                            logger.info(f'MissingValue: -línea {counter}- en -{columna[i]}- no hay tasa especificada - Registro se sacará de la data.\n')
                            errores.append(counter)
                        else: 
                            if columna[i] == 'p_porcentaje_amortizable' and self.df['t_paymentType'][counter] != 'bullet':
                                logger.info(f'MissingValue: -línea {counter}- en -{columna[i]}- no hay tasa especificada - Registro se sacará de la data.\n')
                                errores.append(counter)
                    else: 
                        if type(cell) != float and type(cell) != int:
                            dato = str(cell)
                            if '%' in dato:
                                logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- es tipo {type(cell)} - Registro se convertirá a número.\n')
                                self.df.iloc[counter, i] = float(dato.strip('%')) / 100
                            else:
                                logger.info(f'TypeError: -línea {counter}- en -{columna[i]}- no es un número - Registro se sacará de la data.\n')
                                errores.append(counter)
                        else:
                            if cell < 0:
                                logger.info(f'ValueError: -línea {counter}- en -{columna[i]}- es menor que 0 - Registro se convertirá a positivo.\n')
                                self.df.replace(self.df[columna[i]][counter], -cell, inplace = True)
                            elif cell > 1:
                                logger.info(f'ValueError: -línea {counter}- en -{columna[i]}- es mayor que 1 - Registro se dividirá por 100.\n')
                                self.df.replace(df[columna[i]][counter], cell / 100, inplace = True)
                    counter += 1
            elif columna[i] == 't_paymentType':
                for cell in self.df[columna[i]]:
                    if cell == 'bullet':
                        self.df.iloc[counter, i + 4 : i + 4 + 3] = np.nan
                    counter += 1
            elif columna[i] == 's_couponType':
                if cell == 'normal':
                    self.df.replace(self.df.iloc[counter, i + 1], '', inplace = True)
            
            dicErrores[columna[i]] = errores

        conteo_errores = self.conteoErrores(dicErrores)
        if conteo_errores > 0:
            logger.info(f'AVISO: Validado con {conteo_errores} {"errores" if conteo_errores > 1 else "error"}\n')
        else:
            logger.info('Validado sin errores!\n')

        #Lista de líneas a eliminar
        lineas = []           
        for key, values in dicErrores.items():
            dicErrores = {key: values for key, values in dicErrores.items()}
            if len(values) > 0:
                for i in range(len(values)):
                    lineas.append(values[i])

        logger.info(f'AVISO: Se procederá a eliminar las líneas {lineas} de la data - Se importarán {len(self.df) - len(lineas)} líneas.')
        #self.df = self.df.dropna().reset_index(drop = True)
        self.tiempo_final_calculo = time.perf_counter()
        self.tiempo_calculo = self.tiempo_final_calculo - self.tiempo_inicio_calculo
        
        logger.removeHandler(fileHandler)
        fileHandler.close()

    def getResults(self):

        return self.df

    def processTime(self):

        return self.tiempo_calculo

    def conteoErrores(self, my_dict):
        conteo = 0
        # Create a Counter object for each list in the dictionary
        counters = {key: Counter(values) for key, values in my_dict.items()}
        # Display the counts for each list in the dictionary
        for key, counter in counters.items():
            #print(f"Conteo para '{key}':")
            for element, count in counter.items():
            #    print(f"{element}: {count}")
                conteo += count
        return conteo